/*
 * Support for executing code in new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>
#include <l4/util/util.h>

#include <exec/elf.h>
#include <exec/memory.h>
#include <exec/pager.h>
#include <exec/process.h>
#include <ipc/server.h>

#include <stdio.h>

#include <pthread-l4.h>
#include <pthread.h>

#include "pager_object_server.h"



static ExecPager exec_pager;



int main(int argc, char *argv[])
{
  long err;

  if (argc < 2)
  {
    printf("Need a program to run.\n");
    return 1;
  }

  /* Initialise the memory of the new task. */

  offset_t initial_stack_size = 16 * L4_PAGESIZE;
  ExplicitSegment stack(Utcb_area_start - initial_stack_size, initial_stack_size, L4_FPAGE_RW);
  Payload *payload;

  if (exec_get_payload(argv[1], &payload))
  {
    printf("Could not initialise program.\n");
    return 1;
  }

  if (stack.allocate())
  {
    printf("Could not allocate stack.\n");
    return 1;
  }

  /* Initialise pager regions. */

  for (unsigned int i = 0; i < payload->segments(); i++)
  {
    if (payload->segment(i)->loadable())
      exec_pager.add(payload->segment(i)->region());
  }

  exec_pager.add(stack.region());

  /* Start the pager. */

  pthread_t pager_thread;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  ipc_server_config_type config;
  ipc_server_init_for(&config, PagerObject, &exec_pager);

  err = pthread_create(&pager_thread, &attr, ipc_server_start_mainloop, &config);

  if (err)
  {
    printf("Could not start pager thread.\n");
    return 1;
  }

  printf("Starting pager thread...\n");
  err = ipc_server_start_config_thread(&config, pthread_l4_cap(pager_thread));

  if (err)
  {
    printf("Could not start pager.\n");
    return 1;
  }

  /* Configure the environment for the task, specifying the pager (and exception
     handler plus region mapper). */

  Process process;

  err = process.configure(config.server);

  if (err)
  {
    printf("Could not configure task.\n");
    return 1;
  }

  /* Populate a thread stack with argument and environment details. */

  Stack st(stack);

  /* NOTE: Environment vector is currently not defined. */

  char *envp[] = {NULL};

  st.populate(argc - 1, argv + 1, envp);

  /* Start the new thread in the given stack. */

  printf("Run thread...\n");

  err = process.thread_start(payload->entry_point(), st);

  if (err)
  {
    printf("Could not run thread.\n");
    return 1;
  }

  printf("Finished.\n");
  while (1);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
